/*
 * Copyright (c) 1996-2009 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as "Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "common/h/Types.h"
#include "common/h/Vector.h"
#include "common/h/Dictionary.h"

#include "arch.h"
#include "util.h"
#include "process.h"
#include "symtab.h"
#include "instPoint.h"
#include "InstrucIter.h"
#include "mapped_object.h"

#include "BPatch_Set.h"

#include "BPatch_instruction.h"
#include "BPatch_memoryAccess_NP.h"

#include "registerSpace.h"

/* Prints the op code */
void InstrucIter::printOpCode()
{
  const instruction i = getInstruction();
  printf("OpCode num %d",(*i).xlform.op);
}

bool InstrucIter::isRegConstantAssignment(int * regArray, Address * regWL)
{
  const instruction i = getInstruction();
  int regNum = (*i).dform.rt;
  

  if (isA_RT_WriteInstruction() || isA_RA_WriteInstruction())
  {
    if (regNum >=3 && regNum <= 10)
      regWL[regNum] = current;
  }
  
  // int regNum = (*i).dform.rt;
  if((*i).xlform.op == 14 && (*i).dform.ra == 0 
     && regNum >= 3 && regNum <= 10)
  {
    regArray[regNum]  = (*i).dform.d_or_si;
    //if(regNum == 8)
    //printf("Value for reg 8 is %d\n",regArray[regNum]);
      
    return true;
  }
  else
    return false;
}


bool InstrucIter::isClauseInstruction()
{
  const instruction i = getInstruction();
  if((*i).xlform.op == 14 && (*i).dform.rt == 3 && (*i).dform.ra == 0)
  {
    printf("Clause value is %d\n", (*i).dform.d_or_si);
    return true;
  }
  else
    return false;
}

signed InstrucIter::getDFormDValue()
{
  const instruction i = getInstruction();
  return (*i).dform.d_or_si;
}

/* Returns the value of the RT Register */
unsigned InstrucIter::getRTValue(void)
{
  const instruction i = getInstruction();
  return (*i).xform.rt;  
}

/* Returns the value of the RA Register */
unsigned InstrucIter::getRAValue(void)
{
  const instruction i = getInstruction();
  return (*i).xform.ra;
}

unsigned InstrucIter::getRBValue(void)
{
  const instruction i = getInstruction();
  return (*i).xform.rb;
}

unsigned InstrucIter::getFRTValue(void)
{
  const instruction i = getInstruction();
  return (*i).aform.frt;
}

unsigned InstrucIter::getFRAValue(void)
{
  const instruction i = getInstruction();
  return (*i).aform.fra;
}

unsigned InstrucIter::getFRBValue(void)
{
  const instruction i = getInstruction();
  return (*i).aform.frb;
}

unsigned InstrucIter::getFRCValue(void)
{
  const instruction i = getInstruction();
  return (*i).aform.frc;
}

/* Returns true if the instruction reads/writes the MX (SPR0) SP Register */
bool InstrucIter::isA_MX_Instruction()
{
  const instruction i = getInstruction();

  if (
      ((*i).xform.op == X_EXTENDEDop
       && ( (*i).xform.xo == SREQxop || (*i).xform.xo == SLEQxop ||
	    (*i).xform.xo == SLLIQxop || (*i).xform.xo == SRLIQxop ||
	    (*i).xform.xo == SRLQxop || (*i).xform.xo == SLLQxop ||
	    (*i).xform.xo == SLIQxop || (*i).xform.xo == SRQxop ||
	    (*i).xform.xo == SLQxop || (*i).xform.xo == SRIQxop ||
	    (*i).xform.xo == SRAIQxop || (*i).xform.xo == SREAxop ||
	    (*i).xform.xo == STSXxop || (*i).xform.xo == STSIxop ||
	    (*i).xform.xo == LSCBXxop
	    ))
      ||
      ((*i).xoform.op == XO_EXTENDEDop
       && ( (*i).xoform.xo == DIVxop || (*i).xoform.xo == DIVSxop ||
	    (*i).xoform.xo == MULxop )))
    return true;
  else
    return false;
}



/* This function returns true if the instruction affects the RD Register */
bool InstrucIter::isA_RT_WriteInstruction()
{
  const instruction i = getInstruction();
  if (
      /* XO Form */
      ((*i).xoform.op == XO_EXTENDEDop
       && ( (*i).xoform.xo == CAXxop || (*i).xoform.xo == Axop 
	    || (*i).xoform.xo == AExop || (*i).xoform.xo == NABSxop 
	    || (*i).xoform.xo == ANDCxop || (*i).xoform.xo == DIVWxop
	    || (*i).xoform.xo == SUBFxop || (*i).xoform.xo == SFxop
	    || (*i).xoform.xo == SFExop || (*i).xoform.xo == SFMExop 
	    || (*i).xoform.xo == SFZExop || (*i).xoform.xo == MULHWxop 
	    || (*i).xoform.xo == DIVWUxop || (*i).xoform.xo == MULSxop 
	    || (*i).xoform.xo == NEGxop || (*i).xoform.xo == ABSxop 
	    || (*i).xoform.xo == DIVxop || (*i).xoform.xo == DIVSxop
	    || (*i).xoform.xo == DOZxop || (*i).xoform.xo == MULxop 
	    || (*i).xoform.xo == SFExop || (*i).xoform.xo == SFMExop 
	    || (*i).xoform.xo == SFZExop || (*i).xoform.xo == MULHWUxop )) ||
	
      /* D Form */
      ((*i).dform.op == MULIop || (*i).dform.op == SFIop || (*i).dform.op == DOZIop   ) || 
      ((*i).dform.op >= SIop && (*i).dform.op  <= CAUop) || 
      ((*i).dform.op >= Lop && (*i).dform.op <= LBZUop) || 
      ((*i).dform.op >= LHZop && (*i).dform.op <= LHAUop) || 
	
      /* X Form */
      ((*i).xform.op == X_EXTENDEDop 
       && ( (*i).xform.xo == LBZUXxop || 
            (*i).xform.xo == LBZXxop || 
	    (*i).xform.xo == LHAUXxop  || 
	    (*i).xform.xo == LHAXxop || 
            (*i).xform.xo == LHBRXxop ||
	    (*i).xform.xo == LHZUXxop  || 
            (*i).xform.xo == LHZXxop || 
	    (*i).xform.xo == LWARXxop || 
            (*i).xform.xo == LWBRXxop  || 
	    (*i).xform.xo == LUXxop || 
            (*i).xform.xo == LXxop ||
	    (*i).xform.xo == MFCRxop  ||  
	    (*i).xform.xo == MFMSRxop || 
            (*i).xform.xo == MFSPRxop  || 
	    (*i).xform.xo == CLCSxop  || 
            (*i).xform.xo == MFSRxop || 
	    (*i).xform.xo == MFSRIxop ||
	    (*i).xform.xo == MFSRINxop  || 
            (*i).xform.xo == RACxop ||
            (*i).xform.xo == LWAXxop ||
            (*i).xform.xo == LWAUXxop ||
            (*i).xform.xo == LDXxop ||
            (*i).xform.xo == LDUXxop ||
            (*i).xform.xo == LDARXxop)) ||
      /* dsform */ 
      ((*i).dsform.op == LDop &&
       ((*i).dsform.xo == LWAxop ||
        (*i).dsform.xo == LDxop ||
        (*i).dsform.xo == LDUxop)))
  {
    return true;
  }
  else
  {
    return false;
  }
}  

/* This function returns true if the instruction affects the FRD Register */
bool InstrucIter::isA_FRT_WriteInstruction()
{
  const instruction i = getInstruction();
  if (
      /* A Form */
      ((*i).aform.op == A_FP_EXTENDEDop2 
       && ( (*i).aform.xo == FAxop || (*i).aform.xo == FDxop || 
	    (*i).aform.xo == FMxop || (*i).aform.xo == FMAxop  || 
	    (*i).aform.xo == FMSxop || (*i).aform.xo == FNMAxop ||
	    (*i).aform.xo == FNMSxop || (*i).aform.xo == FSQRTxop || 
	    (*i).aform.xo == FSxop )) ||  

      ((*i).aform.op == A_FP_EXTENDEDop1
       && ( (*i).aform.xo == FDIVSxop || (*i).aform.xo == FMADDSxop || 
	    (*i).aform.xo == FMULSxop || (*i).aform.xo == FMSUBSxop  || 
	    (*i).aform.xo == FNMSUBSxop || (*i).aform.xo == FNMADDSxop ||
	    (*i).aform.xo == FSUBSxop )) ||  
      /* X Form */
      ((*i).xform.op == X_FP_EXTENDEDop 
       && ( (*i).xform.xo == MFFSxop || (*i).xform.xo == FABSxop ||
	    (*i).xform.xo == FCIRZxop  || (*i).xform.xo == FMRxop || 
	    (*i).xform.xo == FNABSxop || (*i).xform.xo == FCIRxop ||
	    (*i).xform.xo == FNEGxop || (*i).xform.xo == FRSPxop )) ||

      ((*i).xform.op == X_EXTENDEDop
       && ( (*i).xform.xo == LFDUXxop  || (*i).xform.xo == LFDXxop || 
	    (*i).xform.xo == LFSUXxop || (*i).xform.xo == LFSXxop )))
  {
    return true;
  }
  else
  {
    return false;
  }
}  


/* This function returns true if the instruction reads the RT Register */
bool InstrucIter::isA_RT_ReadInstruction()
{
  const instruction i = getInstruction();
  if (
      /* D Form */
      ((*i).dform.op >= ORILop && (*i).dform.op  <= ANDIUop) || 
      ((*i).dform.op >= STop && (*i).dform.op  <= STBUop) ||
      (*i).dform.op == STMop ||

      /* M form */
      ((*i).mform.op >= RLIMIop && (*i).mform.op <= RLNMop) ||

      /*  XFX Form */
      ((*i).xfxform.op == X_EXTENDEDop && (*i).xfxform.xo == MTCRFxop ) ||
      
      /* X Form */
      ((*i).xfxform.op == X_EXTENDEDop
       && ( (*i).xform.xo == SLxop || 
            (*i).xform.xo == CNTLZxop ||
	    (*i).xform.xo == ANDxop  || 
            (*i).xform.xo == MASKGxop || 
	    (*i).xform.xo == ANDCxop || 
            (*i).xform.xo == EXTSxop ||
	    (*i).xform.xo == NORxop  || 
            (*i).xform.xo == STWCXxop ||
	    (*i).xform.xo == STXxop  || 
            (*i).xform.xo == SLQxop || 
	    (*i).xform.xo == SLExop ||
	    (*i).xform.xo == STUXxop  ||
            (*i).xform.xo == SLIQxop || 
	    (*i).xform.xo == STBXxop ||
	    (*i).xform.xo == SLLQxop  ||
            (*i).xform.xo == SLEQxop || 
	    (*i).xform.xo == STBUXxop ||
	    (*i).xform.xo == SLLIQxop  || 
            (*i).xform.xo == EQVxop || 
	    (*i).xform.xo == XORxop ||
	    (*i).xform.xo == ORCxop  || 
            (*i).xform.xo == ORxop ||
	    (*i).xform.xo == MTSPRxop  || 
            (*i).xform.xo == NANDxop || 
	    (*i).xform.xo == SRxop ||
	    (*i).xform.xo == RRIBxop  ||
            (*i).xform.xo == MASKIRxop || 
	    (*i).xform.xo == STSXxop ||
	    (*i).xform.xo == STBRXxop  ||
            (*i).xform.xo == SRQxop || 
	    (*i).xform.xo == SRExop ||
	    (*i).xform.xo == SRIQxop  ||
            (*i).xform.xo == STSIxop || 
	    (*i).xform.xo == SRLQxop ||
	    (*i).xform.xo == SREQxop  ||
            (*i).xform.xo == SRLIQxop ||
	    (*i).xform.xo == SRAxop ||
	    (*i).xform.xo == SRAIxop  || 
            (*i).xform.xo == SRAQxop || 
	    (*i).xform.xo == SREAxop ||
	    (*i).xform.xo == EXTSxop ||
            (*i).xform.xo == SRAIQxop || 
	    (*i).xform.xo == EXTSBxop ||
            (*i).xform.xo == STDXxop || (*i).xform.xo == STDUXxop ||
            (*i).xform.xo == STDCXxop)) ||
      ((*i).dsform.op == STDop &&
       ((*i).dsform.xo == STDxop ||
        (*i).dsform.xo == STDUxop)) ||
      ((*i).mdform.op == RLDop))
  {
    return true;
  }
  else
  {
    return false;
  }
}

/* This function returns true if the instruction reads the FRT Register */
bool InstrucIter::isA_FRT_ReadInstruction()
{
  const instruction i = getInstruction();
  if (
      /* D Form */
      ((*i).dform.op == STHop || (*i).dform.op == STHUop ) || 
      ((*i).dform.op >= STFSop && (*i).dform.op  <= STFDUop) ||
      /* X Form */
      ((*i).xform.op == X_EXTENDEDop 
       && ( (*i).xform.xo == STFSXxop || (*i).xform.xo == STFSUXxop  || 
	    (*i).xform.xo == STFDXxop || (*i).xform.xo == STFDUXxop ||  
	    (*i).xform.xo == STFIWXxop )))
  {
    return true;
  }
  else
  {
    return false;
  }
}  

  
/* This function returns true if the instruction reads the RA Register */
bool InstrucIter::isA_RA_ReadInstruction()
{
  const instruction i = getInstruction();
  if (
      /* XO Form */
      ((*i).xoform.op == XO_EXTENDEDop 
       && ( (*i).xoform.xo == SFxop || (*i).xoform.xo == Axop || (*i).xoform.xo == MULHWUxop ||
	    (*i).xoform.xo == SUBFxop  || (*i).xoform.xo == MULHWxop || (*i).xoform.xo == NEGxop ||
	    (*i).xoform.xo == MULxop || (*i).xoform.xo == SFExop || (*i).xoform.xo == AExop || 
	    (*i).xoform.xo == SFZExop || (*i).xoform.xo == AZExop || (*i).xoform.xo == SFMExop ||  
	    (*i).xoform.xo == AMExop || (*i).xoform.xo == DOZxop || (*i).xoform.xo == CAXxop ||
	    (*i).xoform.xo == DIVxop || (*i).xoform.xo == ABSxop || (*i).xoform.xo == DIVSxop ||
	    (*i).xoform.xo == DIVWUxop || (*i).xoform.xo == NABSxop || (*i).xoform.xo == DIVWxop ||
            (*i).xoform.xo == MULSxop)) ||
      /* D Form */
      ((*i).dform.op == TIop || (*i).dform.op == MULIop || 
       (*i).dform.op == SFIop || (*i).dform.op == DOZIop   ) || 
      ((*i).dform.op >= CMPLIop && (*i).dform.op < CALop) ||
      ((*i).dform.op >= CALop && (*i).dform.op <= CAUop && (*i).dform.ra) ||
      ((*i).dform.op >= Lop && (*i).dform.op <= STFDUop) ||
      
      /* X Form */
      ((*i).xform.op == X_EXTENDEDop
       && ( (*i).xform.xo == CMPxop || (*i).xform.xo == Txop ||
	    (*i).xform.xo == LWARXxop  || (*i).xform.xo == LXxop || (*i).xform.xo == CMPLxop ||
	    (*i).xform.xo == DCBSxop  || (*i).xform.xo == LUXxop || (*i).xform.xo == DCBFxop ||
	    (*i).xform.xo == LBZXxop  || (*i).xform.xo == CLFxop || (*i).xform.xo == LBZUXxop ||
	    (*i).xform.xo == STWCXxop  || (*i).xform.xo == STXxop || (*i).xform.xo == STUXxop ||
	    (*i).xform.xo == STBXxop  || (*i).xform.xo == DCBTSTxop || (*i).xform.xo == STBUXxop ||
	    (*i).xform.xo == LSCBXxop  || (*i).xform.xo == DCBTxop || (*i).xform.xo == LHZXxop ||
            (*i).xform.xo == LHZUXxop ||
	    (*i).xform.xo == TLBIxop  || (*i).xform.xo == DIVxop || (*i).xform.xo == LHAXxop ||
	    (*i).xform.xo == LHAUXxop  || (*i).xform.xo == STHXxop || (*i).xform.xo == STHUXxop ||
	    (*i).xform.xo == DCBIxop || (*i).xform.xo == CLIxop || (*i).xform.xo == CLCSxop ||
	    (*i).xform.xo == LSXxop  || (*i).xform.xo == LWBRXxop || (*i).xform.xo == LFSXxop ||
	    (*i).xform.xo == LFSUXxop  || (*i).xform.xo == LSIxop || (*i).xform.xo == LFDXxop ||
	    (*i).xform.xo == DCLSTxop  || (*i).xform.xo == LFDUXxop || (*i).xform.xo == STSXxop ||
	    (*i).xform.xo == STBRXxop  || (*i).xform.xo == STFSXxop || (*i).xform.xo == STFSUXxop ||
	    (*i).xform.xo == STSIxop  || (*i).xform.xo == STFDXxop || (*i).xform.xo == STFDUXxop ||
	    (*i).xform.xo == LHBRXxop  || (*i).xform.xo == RACxop ||
	    (*i).xform.xo == STHBRXxop || (*i).xform.xo == ICBIxop || 
	    (*i).xform.xo == STFIWXxop || (*i).xform.xo == DCLZxop ||
            (*i).xform.xo == LWAXxop || (*i).xform.xo == LWAUXxop ||
            (*i).xform.xo == LDXxop || (*i).xform.xo == LDUXxop ||
            (*i).xform.xo == STDXxop || (*i).xform.xo == STDUXxop ||
            (*i).xform.xo == LDARXxop || (*i).xform.xo == STDCXxop)) ||
      ((*i).dsform.op == LDop &&
       ((*i).dsform.xo == LWAxop ||
        (*i).dsform.xo == LDxop ||
        (*i).dsform.xo == LDUxop)) ||
      ((*i).dsform.op == STDop &&
       ((*i).dsform.xo == STDxop ||
        (*i).dsform.xo == STDUxop)))
  {
    return true;
  }
  else
  {
    return false;
  }
}  


/* This function returns true if the instruction reads the RB Register */
bool InstrucIter::isA_RB_ReadInstruction()
{
  const instruction i = getInstruction();
  if (
      /* XO Form */
      ((*i).xoform.op == XO_EXTENDEDop
       && ( (*i).xoform.xo == SFxop || (*i).xoform.xo == Axop || (*i).xoform.xo == MULHWUxop ||
	    (*i).xoform.xo == SUBFxop  || (*i).xoform.xo == MULHWxop || 
	    (*i).xoform.xo == MULxop || (*i).xoform.xo == SFExop ||
	    (*i).xoform.xo == MULSxop || (*i).xoform.xo == AExop || 
	    (*i).xoform.xo == DOZxop || (*i).xoform.xo == CAXxop ||
	    (*i).xoform.xo == DIVxop || (*i).xoform.xo == DIVSxop ||
	    (*i).xoform.xo == DIVWUxop || (*i).xoform.xo == DIVWxop)) ||
      /* M Form */
      ((*i).mform.op == RLMIop || (*i).mform.op == RLNMop) ||
     
      /* X Form */
      ((*i).xform.op == X_EXTENDEDop
       && ( (*i).xform.xo == CMPxop || (*i).xform.xo == Txop ||
	    (*i).xform.xo == LWARXxop  || (*i).xform.xo == LXxop || (*i).xform.xo == SLxop ||
	    (*i).xform.xo == ANDxop  || (*i).xform.xo == MASKGxop || (*i).xform.xo == CMPLxop ||
	    (*i).xform.xo == DCBSxop  || (*i).xform.xo == LUXxop || (*i).xform.xo == ANDCxop ||
	    (*i).xform.xo == DCBFxop  || (*i).xform.xo == LBZXxop || (*i).xform.xo == CLFxop ||
	    (*i).xform.xo == LBZUXxop  || (*i).xform.xo == NORxop || (*i).xform.xo == STWCXxop ||
	    (*i).xform.xo == STXxop  || (*i).xform.xo == SLQxop || (*i).xform.xo == SLExop ||
	    (*i).xform.xo == STUXxop  || (*i).xform.xo == STBXxop || (*i).xform.xo == SLLQxop ||
	    (*i).xform.xo == SLEQxop  || (*i).xform.xo == DCBTSTxop || (*i).xform.xo == STBUXxop ||
	    (*i).xform.xo == LSCBXxop  || (*i).xform.xo == DCBTxop || (*i).xform.xo == LHZXxop ||
            (*i).xform.xo == LHZUXxop ||
	    (*i).xform.xo == EQVxop  || (*i).xform.xo == TLBIxop || (*i).xform.xo == XORxop ||
	    (*i).xform.xo == DIVxop  || (*i).xform.xo == LHAXxop || (*i).xform.xo == LHAUXxop ||
	    (*i).xform.xo == STHXxop  || (*i).xform.xo == ORCxop || (*i).xform.xo == STHUXxop ||
	    (*i).xform.xo == ORxop || (*i).xform.xo == DCBIxop || (*i).xform.xo == NANDxop ||
	    (*i).xform.xo == CLIxop  || (*i).xform.xo == LSXxop || (*i).xform.xo == LWBRXxop ||
	    (*i).xform.xo == LFSXxop  || (*i).xform.xo == SRxop || (*i).xform.xo == RRIBxop ||
	    (*i).xform.xo == MASKIRxop  || (*i).xform.xo == LFSUXxop || (*i).xform.xo == LFDXxop ||
	    (*i).xform.xo == DCLSTxop  || (*i).xform.xo == LFDUXxop || (*i).xform.xo == STSXxop ||
	    (*i).xform.xo == STBRXxop  || (*i).xform.xo == STFSXxop || (*i).xform.xo == SRQxop ||
	    (*i).xform.xo == SRExop  || (*i).xform.xo == SRIQxop || (*i).xform.xo == STFDXxop ||
	    (*i).xform.xo == SREQxop  || (*i).xform.xo == STFDUXxop || (*i).xform.xo == LHBRXxop ||
	    (*i).xform.xo == SRAxop || (*i).xform.xo == RACxop ||
	    (*i).xform.xo == STHBRXxop ||
	    (*i).xform.xo == SRAQxop  || (*i).xform.xo == SREAxop ||
	    (*i).xform.xo == ICBIxop || (*i).xform.xo == STFIWXxop ||
	    (*i).xform.xo == DCLZxop ||
            (*i).xform.xo == LWAXxop || (*i).xform.xo == LWAUXxop ||
            (*i).xform.xo == LDXxop || (*i).xform.xo == LDUXxop ||
            (*i).xform.xo == STDXxop || (*i).xform.xo == STDUXxop ||
            (*i).xform.xo == LDARXxop || (*i).xform.xo == STDCXxop)))
     
  {
    return true;
  }
  else
  {
    return false;
  }
}  




/* This function returns true if the instruction affects the FRA Register */
bool InstrucIter::isA_FRA_ReadInstruction()
{
  const instruction i = getInstruction();
  if (
      /* A Form */
      ((*i).aform.op == A_FP_EXTENDEDop2 
       && ( (*i).aform.xo == FDxop || (*i).aform.xo == FSxop || (*i).aform.xo == FAxop ||
	    (*i).aform.xo == FMxop  || (*i).aform.xo == FMSxop || (*i).aform.xo == FMAxop ||
	    (*i).aform.xo == FNMSxop || (*i).aform.xo == FNMAxop )) ||

      ((*i).aform.op == A_FP_EXTENDEDop1
       && ( (*i).aform.xo == FDIVSxop || (*i).aform.xo == FSUBSxop || 
	    (*i).aform.xo == FADDSxop || (*i).aform.xo == FMULSxop  || 
	    (*i).aform.xo == FMSUBSxop || (*i).aform.xo == FMADDSxop ||
	    (*i).aform.xo == FNMSUBSxop || (*i).aform.xo == FNMADDSxop )) ||  
      
      /* XL Form */
      ((*i).xlform.op == X_FP_EXTENDEDop && (*i).xlform.xo == FCMPUxop ) || 
      /* X Form */
      ((*i).xform.op == X_FP_EXTENDEDop && (*i).xform.xo == FCMPOxop ))

  {
    return true;
  }
  else
  {
    return false;
  }
}  

/* This function returns true if the instruction affects the FRB Register */
bool InstrucIter::isA_FRB_ReadInstruction()
{
  const instruction i = getInstruction();
  if (
      /* A Form */
      ((*i).aform.op == A_FP_EXTENDEDop2 
       && (  (*i).aform.xo == FDxop || (*i).aform.xo == FSxop || 
	     (*i).aform.xo == FAxop ||
	     (*i).aform.xo == FSQRTxop ||
	     (*i).aform.xo == FMxop  || (*i).aform.xo == FMSxop || 
	     (*i).aform.xo == FMAxop ||
	     (*i).aform.xo == FNMSxop || (*i).aform.xo == FNMAxop )) ||

      ((*i).aform.op == A_FP_EXTENDEDop1
       && ( (*i).aform.xo == FDIVSxop || (*i).aform.xo == FSUBSxop || 
	    (*i).aform.xo == FADDSxop ||
	    (*i).aform.xo == FMULSxop  || (*i).aform.xo == FMSUBSxop || 
	    (*i).aform.xo == FMAxop ||
	    (*i).aform.xo == FNMSxop || (*i).aform.xo == FNMAxop )) ||  
      
      /* XL Form */
      ((*i).xlform.op == X_EXTENDEDop 
       && (*i).xlform.xo == FCMPUxop ) || 
     

      /* X Form */
      ((*i).xform.op == X_EXTENDEDop 
       && ( (*i).xform.xo == FRSPxop || (*i).xform.xo == FCIRxop || 
	    (*i).xform.xo == FCIRZxop || (*i).xform.xo == FCMPOxop || 
	    (*i).xform.xo == FNEGxop ||
	    (*i).xform.xo == FMRxop ||
	    (*i).xform.xo == FNABSxop || (*i).xform.xo == FABSxop )))
  {
    return true;
  }
  else
  {
    return false;
  }
}  

/* This function returns true if the instruction reads the FRC Registers */
bool InstrucIter::isA_FRC_ReadInstruction()
{
  const instruction i = getInstruction();
  if (
      /* A Form */
      ((*i).aform.op == A_FP_EXTENDEDop2
       && ( (*i).aform.xo == FMxop || (*i).aform.xo == FMSxop || (*i).aform.xo == FMAxop ||
	    (*i).aform.xo == FNMSxop  || (*i).aform.xo == FNMAxop )))
    return true;
  else
    return false;
	   
}


/* This function returns true if the instruction affects the FRA Register */
bool InstrucIter::isA_FRA_WriteInstruction()
{
  const instruction i = getInstruction();
  if (
      /* A Form */
      ((*i).aform.op == A_FP_EXTENDEDop2
       && ( (*i).aform.xo == FAxop || (*i).aform.xo == FDxop || (*i).aform.xo == FMxop ||
	    (*i).aform.xo == FMAxop  || (*i).aform.xo == FMSxop || (*i).aform.xo == FNMAxop ||
	    (*i).aform.xo == FNMSxop || (*i).aform.xo == FSxop )) ||  
      
      ((*i).aform.op == A_FP_EXTENDEDop1
       && ( (*i).aform.xo == FDIVSxop || (*i).aform.xo == FMADDSxop ||
	    (*i).aform.xo == FMULSxop ||
	    (*i).aform.xo == FMSUBSxop  || (*i).aform.xo == FNMSUBSxop 
	    || (*i).aform.xo == FNMADDSxop ||
	    (*i).aform.xo == FSUBSxop )) ||  
      
      /* X Form */
      ((*i).xform.op == X_FP_EXTENDEDop 
       && ((*i).xform.xo == FCMPOxop || (*i).xform.xo == FCMPUxop )))

  {
    return true;
  }
  else
  {
    return false;
  }
}  



/* This function returns true if the instruction affects the RA Register */
bool InstrucIter::isA_RA_WriteInstruction()
{
  const instruction i = getInstruction();
  if (
      /* X Form */
      ( (*i).xform.op == X_EXTENDEDop
	&& ( (*i).xform.xo == NORxop ||  (*i).xform.xo == STBUXxop || (*i).xform.xo == ANDxop ||
	     (*i).xform.xo == ORxop  || (*i).xform.xo == ORCxop || (*i).xform.xo == SLxop ||
	     (*i).xform.xo == ANDCxop  || (*i).xform.xo == EQVxop ||
	     (*i).xform.xo == EXTSBxop  || (*i).xform.xo == EXTSxop  || (*i).xform.xo == LHZUXxop  ||
	     (*i).xform.xo == LUXxop   || (*i).xform.xo == LBZUXxop  || (*i).xform.xo == LHAUXxop ||
	     (*i).xform.xo == NANDxop   || (*i).xform.xo == STHUXxop   || (*i).xform.xo == STUXxop  ||
	     (*i).xform.xo == XORxop  || (*i).xform.xo == SRQxop   || (*i).xform.xo == SRxop   ||  
	     (*i).xform.xo == CLFxop   || (*i).xform.xo == CLIxop   || (*i).xform.xo == DCLSTxop  ||
	     (*i).xform.xo == MASKGxop   || (*i).xform.xo == MASKIRxop   || (*i).xform.xo == MTSRIxop  ||
	     (*i).xform.xo == RRIBxop   || (*i).xform.xo == SLxop   || (*i).xform.xo == SLExop  ||
	     (*i).xform.xo == SLEQxop   || (*i).xform.xo == SLIQxop   || (*i).xform.xo == SLLIQxop  ||
	     (*i).xform.xo == SLLQxop   || (*i).xform.xo == SLQxop   || (*i).xform.xo == SRAIQxop  ||
	     (*i).xform.xo == SRAQxop   || (*i).xform.xo == SRAxop   || (*i).xform.xo == SRAIxop  ||
	     (*i).xform.xo == SRExop   || (*i).xform.xo == SREAxop   || (*i).xform.xo == SREQxop  ||
	     (*i).xform.xo == LFDUXxop   || (*i).xform.xo == LFDXxop   ||
	     (*i).xform.xo == LFSUXxop   || (*i).xform.xo == LFSXxop  ||
	     (*i).xform.xo == SRIQxop   || (*i).xform.xo == SRLIQxop   || (*i).xform.xo == SRLQxop  ||
	     (*i).xform.xo == SRAxop  || (*i).xform.xo == SRAIxop || (*i).xform.xo == SRxop ||
             (*i).xform.xo == LWAUXxop || (*i).xform.xo == LDUXxop ||
             (*i).xform.xo == STDUXxop)) ||
      
      /* D Form */
      ((*i).dform.op == ANDILop || (*i).dform.op == ANDIUop || (*i).dform.op == LBZUop 
       || (*i).dform.op == LHAUop ||(*i).dform.op == LHZUop || (*i).dform.op == LUop || 
       (*i).dform.op == ORILop || (*i).dform.op == ORIUop || (*i).dform.op == STBUop || 
       (*i).dform.op == STHUop || (*i).dform.op == STUop || (*i).dform.op == XORILop ||
       (*i).dform.op == XORIUop || (*i).dform.op == SFIop || (*i).dform.op == SIop) || 
      
      /* M Form */
      ((*i).mform.op == RLIMIop || (*i).mform.op == RLINMxop || (*i).mform.op == RLMIop 
       ||  (*i).mform.op == RLNMop) ||
      /* DS form */
      ((*i).dsform.op == LDop &&
       ((*i).dsform.xo == LDUxop)) ||
      ((*i).dsform.op == STDop &&
       ((*i).dsform.xo == STDUxop)) ||
      (*i).mdform.op == RLDop)
  {
    return true;
  }
  else
  {
    return false;
  }
  
}

/* This function returns true if the instruction reads from multiple RT Registers */
bool InstrucIter::isA_MRT_ReadInstruction()
{
  const instruction i = getInstruction();
  if (
      /* D Form */
      ((*i).dform.op == STMop) ||
      /* X Form */
      ((*i).xform.op == X_EXTENDEDop
       && ((*i).xform.xo == STSIxop || (*i).xform.xo == STSXxop)))
  {
    return true;
  }
  else
  {
    return false;
  }
}

bool InstrucIter::isA_MRT_WriteInstruction()
{
  const instruction i = getInstruction();
  if (
      /* X Form */
      ((*i).dform.op == LMop) ||
      /* X Form */
      ((*i).xform.op == X_EXTENDEDop
       && ((*i).xform.xo == LSIxop || (*i).xform.xo == LSXxop)))
  {
    return true;
  }
  else
  {
    return false;
  }
}


//some more function used to identify the properties of the instruction
/** is the instruction used to return from the functions
 * @param i the instruction value 
 */
bool InstrucIter::isAReturnInstruction()
{
  const instruction i = getInstruction();
  if(((*i).xlform.op == BCLRop) &&
     ((*i).xlform.xo == BCLRxop) && 
     ((*i).xlform.bt & 0x10) && ((*i).xlform.bt & 0x4) &&
     !((*i).xlform.bb) 
#if defined(os_linux)   // not sure if this is safe on AIX, yet
     // Bernat, 9JUL07. A linking branch-to-link-register is
     // unlikely to be a return instruction.
     //
     // nater 16FEB09 unless the instruction hint indicates subroutine return
     //               (BH is low 2 bits of BB in XL-FORM)
     && (!((*i).xlform.lk) || !((*i).xlform.bb & 3))
#endif
     )
    return true;
  return false;
}


/** is the instruction used to return from the functions,
    dependent upon a condition register
    * @param i the instruction value 
    */
bool InstrucIter::isACondReturnInstruction()
{
  const instruction i = getInstruction();
  if(((*i).xlform.op == BCLRop) &&
     ((*i).xlform.xo == BCLRxop) &&
     ((*i).xlform.bt & 0x14) != 0x14)
    return true;
  return false;
}

/** is the instruction an indirect jump instruction 
 * @param i the instruction value 
 */
bool InstrucIter::isAIndirectJumpInstruction()
{
  const instruction i = getInstruction();
  if(((*i).xlform.op == BCLRop) && ((*i).xlform.xo == BCCTRxop) &&
     !(*i).xlform.lk && ((*i).xlform.bt & 0x10) && ((*i).xlform.bt & 0x4))
    return true;
  if(((*i).xlform.op == BCLRop) && ((*i).xlform.xo == BCLRxop) &&
     !(*i).xlform.lk && ((*i).xlform.bt & 0x10) && ((*i).xlform.bt & 0x4) &&
     ((*i).xlform.bb == 0x1))
    return true;
#if 0
  // Disabled; a branch to link register is a return until we can prove
  // otherwise, not a indirect branch.
  if(((*i).xlform.op == BCLRop) && ((*i).xlform.xo == BCLRxop) &&
     ((*i).xlform.bt & 0x10) && ((*i).xlform.bt & 0x4)){
    if(!ah.hasPrev())
      return false;
    --ah;
    if(!ah.hasPrev())
      return false;
    instruction j = ah.getInstruction();
    if(((*j).xfxform.op == STXop) && ((*j).xfxform.xo == MTSPRxop) &&
       ((*j).xfxform.spr == 0x100)) // LR
      return true;
    ++ah;
    j = ah.getInstruction();
    if(((*j).xfxform.op == STXop) && ((*j).xfxform.xo == MTSPRxop) &&
       ((*j).xfxform.spr == 0x100)) // LR
      return true;
  }
#endif
  return false;
}

/** is the instruction a conditional branch instruction 
 * @param i the instruction value 
 */ 
bool InstrucIter::isACondBranchInstruction()
{
  const instruction i = getInstruction();
  if(((*i).bform.op == BCop) && !(*i).bform.lk &&
     !(((*i).bform.bo & 0x10) && ((*i).bform.bo & 0x4)))
    return true;
  return false;
}

bool InstrucIter::isACondBDZInstruction()
{
  const instruction i = getInstruction();
  if( (*i).bform.op == BCop  && !(*i).bform.lk &&
      ( (*i).bform.bo == 18 || (*i).bform.bo == 19 ||
	(*i).bform.bo == 26 || (*i).bform.bo == 27 ) ) 
    return true;
  return false;
}

bool InstrucIter::isACondBDNInstruction()
{
  const instruction i = getInstruction();
  if( (*i).bform.op == BCop  && !(*i).bform.lk &&
      ( (*i).bform.bo == 16 || (*i).bform.bo == 17 ||
	(*i).bform.bo == 24 || (*i).bform.bo == 25 ) ) 
    return true;
  return false;
}


/** is the instruction an unconditional branch instruction 
 * @param i the instruction value 
 */
bool InstrucIter::isAJumpInstruction()
{
  const instruction i = getInstruction();
  if(((*i).iform.op == Bop) && !(*i).iform.lk)
    return true;
  if(((*i).bform.op == BCop) && !(*i).bform.lk &&
     ((*i).bform.bo & 0x10) && ((*i).bform.bo & 0x4))
    return true;
  return false;
}
/** is the instruction a call instruction 
 * @param i the instruction value 
 */
bool InstrucIter::isACallInstruction()
{

  //cerr << current << endl;
  const instruction i = getInstruction();
  //const instruction i = getInstruction();
#define CALLmatch 0x48000001 /* bl */
        
  // Only look for 'bl' instructions for now, although a branch
  // could be a call function, and it doesn't need to set the link
  // register if it is the last function call
  //cerr << currentAddreOPmask << endl;

  return (i.isInsnType(OPmask | LLmask, CALLmatch));
  /*
    const instruction i = getInstruction();
    if((*i).iform.lk && 
    (((*i).iform.op == Bop) || ((*i).bform.op == BCop) ||
    (((*i).xlform.op == BCLRop) && 
    (((*i).xlform.xo == 16) || ((*i).xlform.xo == 528))))){
    return true;
    }
    return false;

  */
}
bool InstrucIter::isADynamicCallInstruction()
{
  // I'm going to break this up a little so that I can comment
  // it better. 
    
  const instruction i = getInstruction();

  if ((*i).xlform.op == BCLRop && (*i).xlform.xo == BCLRxop)
  {
    if ((*i).xlform.lk)
      // Type one: Branch-to-LR, save PC in LR
      // Standard function pointer call
      return true;
    else
      // Type two: Branch-to-LR, don't save PC
      // Haven't seen one of these around
      // It would be a return instruction, probably
    {
      return false;
    }
  }
  if ((*i).xlform.op == BCLRop && (*i).xlform.xo == BCCTRxop)
  {
    if ((*i).xlform.lk)
      // Type three: Branch-to-CR, save PC
    {
      return true;
    }
    else
      // Type four: Branch-to-CR, don't save PC
      // Used for global linkage code.
      // We handle those elsewhere.
    {
      return true;
    }
  }
  return false;


}


bool InstrucIter::isAnneal(){
  return true;
}

#define MK_LD1(bytes, i, imm, ra) (new BPatch_memoryAccess(i, current,true, false, (bytes), (imm), (ra), -1))
#define MK_SD1(bytes, i, imm, ra) (new BPatch_memoryAccess(i, current, false, true, (bytes), (imm), (ra), -1))

#define MK_LX1(bytes, i, ra, rb) (new BPatch_memoryAccess(i, current,true, false, (bytes), 0, (ra), (rb)))
#define MK_SX1(bytes, i, ra, rb) (new BPatch_memoryAccess(i, current,false, true, (bytes), 0, (ra), (rb)))

#define MK_LD(bytes, i) (MK_LD1((bytes), i, (**i).dform.d_or_si, (signed)(**i).dform.ra))
#define MK_SD(bytes, i) (MK_SD1((bytes), i, (**i).dform.d_or_si, (signed)(**i).dform.ra))

// VG(11/20/01): X-forms ignore ra if 0, but not rb...
#define MK_LX(bytes, i) (MK_LX1((bytes), i, ((**i).xform.ra ? (signed)(**i).xform.ra : -1), (**i).xform.rb))
#define MK_SX(bytes, i) (MK_SX1((bytes), i, ((**i).xform.ra ? (signed)(**i).xform.ra : -1), (**i).xform.rb))

#define MK_LDS(bytes, i) (MK_LD1((bytes), i, ((**i).dsform.ds << 2), (signed)(**i).dsform.ra))
#define MK_SDS(bytes, i) (MK_SD1((bytes), i, ((**i).dsform.ds << 2), (signed)(**i).dsform.ra))

#define MK_LI(bytes, i) (MK_LX1((bytes), i, (**i).xform.ra, -1))
#define MK_SI(bytes, i) (MK_SX1((bytes), i, (**i).xform.ra, -1))

struct opCodeInfo {
    static bool doneInit;
  unsigned int bytes; //: 4;
  unsigned int direc; //: 1; // 0 == load, 1 == store
public:
  opCodeInfo(unsigned b, unsigned d) : bytes(b), direc(d) {}
  opCodeInfo() : bytes(0), direc(0) {}
};

bool opCodeInfo::doneInit = false;
opCodeInfo *xopCodes[1024];

void initOpCodeInfo()
{
    opCodeInfo::doneInit = true;
  memset(xopCodes, sizeof(opCodeInfo *)*1024, 0);
  xopCodes[LWARXxop]	= new opCodeInfo(4, 0);
  xopCodes[LDXxop] 	= new opCodeInfo(8, 0);
  xopCodes[LXxop]	= new opCodeInfo(4, 0);
  xopCodes[LDUXxop]	= new opCodeInfo(8, 0);
  xopCodes[LUXxop] 	= new opCodeInfo(4, 0);
  xopCodes[LDARXxop]	= new opCodeInfo(8, 0);
  xopCodes[LBZXxop]	= new opCodeInfo(1, 0);
  xopCodes[LBZUXxop]	= new opCodeInfo(1, 0);

  xopCodes[STDXxop]	= new opCodeInfo(8, 1);
  xopCodes[STWCXxop]	= new opCodeInfo(4, 1);
  xopCodes[STXxop]	= new opCodeInfo(4, 1);
  xopCodes[STDUXxop]	= new opCodeInfo(8, 1);
  xopCodes[STUXxop]	= new opCodeInfo(4, 1);
  xopCodes[STDCXxop]	= new opCodeInfo(8, 1);
  xopCodes[STBXxop]	= new opCodeInfo(1, 1);
  xopCodes[STBUXxop]	= new opCodeInfo(1, 1);

  xopCodes[LHZXxop]	= new opCodeInfo(2, 0);
  xopCodes[LHZUXxop]	= new opCodeInfo(2, 0);
  xopCodes[LHAXxop]	= new opCodeInfo(2, 0);
  xopCodes[LWAXxop]	= new opCodeInfo(4, 0);
  xopCodes[LWAUXxop]	= new opCodeInfo(4, 0);
  xopCodes[LHAUXxop]	= new opCodeInfo(2, 0);

  xopCodes[STHXxop]	= new opCodeInfo(2, 1);
  xopCodes[STHUXxop]	= new opCodeInfo(2, 1);

  xopCodes[LSXxop]	= new opCodeInfo(0, 0); // count field only available at runtime
  xopCodes[LWBRXxop]	= new opCodeInfo(4, 0);
  xopCodes[LFSXxop]	= new opCodeInfo(4, 0);
  xopCodes[LFSUXxop]	= new opCodeInfo(4, 0);
  xopCodes[LSIxop]	= new opCodeInfo(0, 0); // count field needs to be computed
  xopCodes[LFDXxop]	= new opCodeInfo(8, 0);
  xopCodes[LFDUXxop]	= new opCodeInfo(8, 0);

  xopCodes[STSXxop]	= new opCodeInfo(0, 1); // count field only available at runtime
  xopCodes[STBRXxop]	= new opCodeInfo(4, 1);
  xopCodes[STFSXxop]	= new opCodeInfo(4, 1);
  xopCodes[STFSUXxop]	= new opCodeInfo(4, 1);
  xopCodes[STSIxop]	= new opCodeInfo(0, 1); // bytes field needs to be computed
  xopCodes[STFDXxop]	= new opCodeInfo(8, 1);
  xopCodes[STFDUXxop]	= new opCodeInfo(8, 1);

  xopCodes[LHBRXxop]	= new opCodeInfo(2, 0);

  xopCodes[STHBRXxop]	= new opCodeInfo(2, 1);

  xopCodes[STFIWXxop]	= new opCodeInfo(4, 1);

  //bperr( "POWER opcode info initialized.\n");
}


/* return NULL if instruction is not load/store, else returns
   a class that contains info about the load/store */

// VG (09/11/01): I tried to optimize this somewhat. Not sure how much...
// It is a hard-coded binary search that assumes all opcodes have the same
// probability to appear. Probably not realistic, but I don't have statistics
// for Power... Due to the interleaving of load and store opcodes on Power,
// it seems cheaper to have only one function to decide if it is load or store
// Alternatively one could write this based on bits in opcode. E.g. if bit 3
// is 0 then the instruction is most likely a load, else it is likely a store.
// I also assume that no invalid instructions occur: e.g.: LU with RA=0, etc.

#define logIS_A(x)
//#define logIS_A(x) logLine((x))

BPatch_memoryAccess* InstrucIter::isLoadOrStore()
{
  instruction *i = getInsnPtr();

  int op = (**i).dform.op;
  int  b;

  if(op > LXop) { // try D-forms
    if(op < STHop)
      if(op < STop) {
	logIS_A("IS_A: l-lbzu");
	b = op < LBZop ? 4 : 1;
	return MK_LD(b, i);
      } else if (op < LHZop) {
	logIS_A("IS_A: st-stbu");
	b = op < STBop ? 4 : 1;
	return MK_SD(b, i);
      } else {
	logIS_A("IS_A: lhz-lhau");
	return MK_LD(2, i);
      }
    else if(op < STFSop)
      if(op < LFSop)
	if(op < LMop) {
	  logIS_A("IS_A: sth-sthu");
	  return MK_SD(2, i);
	}
	else if(op < STMop) {
	  logIS_A("IS_A: lm");
	  return MK_LD((32 - (**i).dform.rt)*4, i);
	}
	else {
	  logIS_A("IS_A: stm");
	  return MK_SD((32 - (**i).dform.rt)*4, i);
	}
      else {
	logIS_A("IS_A: lfs-lfdu");
	b = op < LFDop ? 4 : 8;
	return MK_LD(b, i);
      }
    else if(op <= STFDUop) {
      logIS_A("IS_A: stfs-stfdu");
      b = op < STFDop ? 4 : 8;
      return MK_SD(b, i);
    }
    else if(op == LDop) {
      logIS_A("IS_A: ld-lwa");
      b = (**i).dsform.xo < 2 ? 8 : 4;
      assert((**i).dsform.xo < 3);
      return MK_LDS(b, i);
    }
    else if(op == STDop) {
      logIS_A("IS_A: std-stdu");
      assert((**i).dsform.xo < 2);
      
      

      return MK_SDS(8, i);
    }
    else
      return BPatch_memoryAccess::none;
  }
  else if(op == LXop) { // X-forms
    unsigned int xop = (**i).xform.xo;
    
    //logIS_A("XOP::%d\n", xop);
    if(!opCodeInfo::doneInit)
        initOpCodeInfo();
    opCodeInfo *oci = xopCodes[xop];

    if (oci == NULL)
      return BPatch_memoryAccess::none;
    else if(oci->bytes > 0)
      return oci->direc ? MK_SX(oci->bytes, i) : MK_LX(oci->bytes, i);
    else if(xop == LSIxop || xop == STSIxop) {
      b = (**i).xform.rb == 0 ? 32 : (**i).xform.rb; 
      return oci->direc ? MK_SI(b, i) : MK_LI(b, i);
    }
    else if(xop == LSXxop || xop == STSXxop) {
      return new BPatch_memoryAccess(i, current,
				     oci->direc == 0, oci->direc == 1,
                                     (long)0, (**i).xform.ra, (**i).xform.rb,
                                     (long)0, POWER_XER2531, -1); // 9999 == XER_25:31
    }
  }
  free(i);
  return BPatch_memoryAccess::none;
}

/** function which returns the offset of control transfer instructions
 * @param i the instruction value 
 */
Address InstrucIter::getBranchTargetAddress(bool *isAbsolute)
{
  const instruction i = getInstruction();
  Address ret = 0;
  if(((*i).iform.op == Bop) || ((*i).bform.op == BCop)){
    int disp = 0;
    if((*i).iform.op == Bop)
      disp = (*i).iform.li;
    else if((*i).bform.op == BCop)
      disp = (*i).bform.bd;
    disp <<= 2;
    if((*i).iform.aa) {
      ret = (Address)disp;
      if (isAbsolute) *isAbsolute = true;
    }
    else
      ret = (Address)(current+disp);
  }
  return (Address)ret;
}

bool InstrucIter::getMultipleJumpTargets(BPatch_Set<Address>& result)
{
  Address initialAddress = current;
  Address TOC_address = 0;
  unsigned TOC_register = 2;
  
  image* img = dynamic_cast<image*>(instructions_);
  process* proc = dynamic_cast<process*>(instructions_);
    
  if (img) { 
    TOC_address = img->getObject()->getTOCoffset();
  }
  else {
    pdvector<mapped_object *> m_objs = proc->mappedObjects();
    for (unsigned i = 0; i < m_objs.size(); i++) {
      void *ptr = m_objs[i]->getPtrToInstruction(current);
      if (ptr) {
	TOC_address = m_objs[i]->parse_img()->getObject()->getTOCoffset();
	TOC_address += m_objs[i]->dataBase();
	break;
      }
    }
  }

  instruction check;
    
  // If there are no prior instructions then we can't be looking at a
  // jump through a jump table.
  if( !hasPrev() ) {
    result += (initialAddress + instruction::size());
    setCurrentAddress(initialAddress);
    return false;
  }
    
  // Check if the previous instruction is a move to CTR or LR;
  // if it is, then this is the pattern we're familiar with.  The
  // register being moved into CTR or LR has the address to jump to.
  (*this)--;
  int jumpAddrReg;
  check = getInstruction();
  if ((*check).xfxform.op == STXop && (*check).xfxform.xo == MTSPRxop &&
      ((*check).xfxform.spr == 0x100 || (*check).xfxform.spr == 0x120)) {
    jumpAddrReg = (*check).xfxform.rt;
    //fprintf(stderr, "Jump table candidate, jump target register is %d\n", jumpAddrReg);
  } else {
    result += (initialAddress + instruction::size());
    setCurrentAddress(initialAddress);
    return false;
  }
  // In the pattern we've seen, if the instruction previous to this is
  // an add with a result that ends up being used as the jump address,
  // then we're adding a relative value we got from the table to a base
  // address to get the jump address; in other words, the contents of
  // the jump table are relative.
  bool tableIsRelative = false;
  if (hasPrev()) {
    (*this)--;
    check = getInstruction();
    if ((*check).xoform.op == CAXop && (*check).xoform.xo == CAXxop &&
	(*check).xoform.rt == (unsigned)jumpAddrReg) {
      tableIsRelative = true; 
      //fprintf(stderr, "table is relative...\n");
    }
    else
      (*this)++;
  }

  setCurrentAddress(initialAddress);    
    
  Address jumpStartAddress = 0;
  Address adjustEntry = 0;
  Address tableStartAddress = 0;
  if(!TOC_address)
  {
    while( hasPrev() ){
      check = getInstruction();
      if((*check).dform.op == CALop) {
         unsigned jumpAddressReg = (*check).dform.ra;
	tableStartAddress = (*check).dform.d_or_si;
	(*this)--;
	check = getInstruction();
	if((*check).dform.op == CAUop && (*check).dform.rt == jumpAddressReg)
	{
	  tableStartAddress += ((*check).dform.d_or_si * 0x10000) & 0xFFFF0000;
	  //fprintf(stderr, "Found CAU and CAL at %lx, ", current);
	  //fprintf(stderr, "tableStartAddress is %lx, high half is %lx\n", tableStartAddress, (*check).dform.d_or_si);
	  break;
	}
	else
	{
	  tableStartAddress = 0;
	}
      }
      else if((*check).dform.op == CAUop) {
         unsigned jumpAddressReg = (*check).dform.rt;
	tableStartAddress = ((*check).dform.d_or_si * 0x10000) & 0xFFFF0000;
	(*this)--;
	check = getInstruction();
	if((*check).dform.op == CALop && (*check).dform.ra == jumpAddressReg)
	{
	  tableStartAddress += (*check).dform.d_or_si;
	  //fprintf(stderr, "Found CAU and CAL at %x, ", current);
	  //fprintf(stderr, "tableStartAddress is %x, low half is %x\n", tableStartAddress, (*check).dform.d_or_si);
	  break;
	}
	else
	{
	  tableStartAddress = 0;
	}
      }
      (*this)--;
    }
    jumpStartAddress = tableStartAddress;
  }
  else if (tableIsRelative) {
    while( hasPrev() ){
      check = getInstruction();
      if(((*check).dform.op == Lop) && ((*check).dform.ra == TOC_register)) { // 32-bit form
	//fprintf(stderr, "Jump offset from TOC: %d\n", (*check).dform.d_or_si);
	jumpStartAddress = 
	(Address)(TOC_address + (*check).dform.d_or_si);
	break;
      }

      if (((*check).dsform.op == LDop) && ((*check).dsform.ra == TOC_register)) { // 64-bit form
	//fprintf(stderr, "Jump offset from TOC: %d\n", ((*check).dsform.ds << 2));
	jumpStartAddress = 
	(Address)(TOC_address + ((*check).dsform.ds << 2));
	break;
      }
      (*this)--;
    }
    // Anyone know what this does?
    (*this)--;
    check = getInstruction();
    if(((*check).dform.op == Lop)) { // 32-bit form
      adjustEntry = (*check).dform.d_or_si;
      //fprintf(stderr, "adjustEntry is 0x%lx (%d)\n",
      //adjustEntry, adjustEntry);
    }
       
    while(hasPrev()){
      instruction check = getInstruction();
      if(((*check).dform.op == Lop) && ((*check).dform.ra == TOC_register)){ // 32-bit form
	//fprintf(stderr, "Table offset from TOC: %d\n", (*check).dform.d_or_si);
	tableStartAddress = 
	(Address)(TOC_address + (*check).dform.d_or_si);
	//fprintf(stderr, "tableStartAddr is 0x%lx\n", tableStartAddress);
	break;
      }
      if(((*check).dsform.op == LDop) && ((*check).dsform.ra == TOC_register)){ // 64-bit form
	//fprintf(stderr, "Table offset from TOC: %d\n", ((*check).dsform.ds << 2));
	tableStartAddress = 
	(Address)(TOC_address + ((*check).dsform.ds << 2));
	//fprintf(stderr, "tableStartAddr is 0x%lx\n", tableStartAddress);
	break;
      }
      (*this)--;
    }
  } else {
    bool foundAdjustEntry = false;
    while( hasPrev() ){
      check = getInstruction();
      if((*check).dform.op == CALop &&
	 (*check).dform.rt == (unsigned)jumpAddrReg &&
	 !foundAdjustEntry){
	foundAdjustEntry = true;
	adjustEntry = (*check).dform.d_or_si;
	jumpAddrReg = (*check).dform.ra;
      } else if((*check).dform.op == Lop &&
		(*check).dform.ra == TOC_register &&
		(*check).dform.rt == (unsigned)jumpAddrReg){
	tableStartAddress = 
	(Address)(TOC_address + (*check).dform.d_or_si);
	break;
      }
      (*this)--;
    }
  }

  // We could also set this = jumpStartAddress...
  if (tableStartAddress == 0)  {
    //fprintf(stderr, "No table start addr, returning\n"); 
    setCurrentAddress(initialAddress);
    return false;
  }

  setCurrentAddress(initialAddress);
  int maxSwitch = 0;
    
  while( hasPrev() ){
    instruction check = getInstruction();
    if(((*check).bform.op == BCop) && 
       !(*check).bform.aa && !(*check).bform.lk){
      (*this)--;
      check = getInstruction();
      if(10 != (*check).dform.op)
	break;
      maxSwitch = (*check).dform.d_or_si + 1;
      break;
    }
    (*this)--;
  }
//fprintf(stderr, "After checking: max switch %d\n", maxSwitch);
  if(!maxSwitch){
    result += (initialAddress + instruction::size());
    //fprintf(stderr, "No maximum, returning\n");
    setCurrentAddress(initialAddress);
    return false;
  }

  Address jumpStart = 0;
  Address tableStart = 0;
  bool is64 = (instructions_->getAddressWidth() == 8);

  if(TOC_address)
  {
    if (img) {
      if (tableIsRelative) {
	void *jumpStartPtr = img->getPtrToData(jumpStartAddress);
	//fprintf(stderr, "jumpStartPtr (0x%lx) = %p\n", jumpStartAddress, jumpStartPtr);
	if (jumpStartPtr)
	  jumpStart = (is64
		       ? *((Address  *)jumpStartPtr)
		       : *((uint32_t *)jumpStartPtr));
	//fprintf(stderr, "jumpStart 0x%lx, initialAddr 0x%lx\n",
	//	jumpStart, initialAddress);
	if (jumpStartPtr == NULL) {
	  setCurrentAddress(initialAddress);
	  return false;
	}
      }
      void *tableStartPtr = img->getPtrToData(tableStartAddress);
      //fprintf(stderr, "tableStartPtr (0x%lx) = %p\n", tableStartAddress, tableStartPtr);
	tableStart = *((Address *)tableStartPtr);
      if (tableStartPtr)
	tableStart = (is64
		      ? *((Address  *)tableStartPtr)
		      : *((uint32_t *)tableStartPtr));
      else {
	setCurrentAddress(initialAddress);                    
	return false;
      }
      //fprintf(stderr, "... tableStart 0x%lx\n", tableStart);
      
      // We're getting an absolute out of the TOC. Figure out
      // whether we're in code or data.
      const fileDescriptor &desc = img->desc();
      Address textStart = desc.code();
      Address dataStart = desc.data();
      
      // I think this is valid on ppc64 linux.  dataStart and codeStart can be 0.
      // assert(jumpStart < dataStart);
      
      bool tableData = false;
      
      if (proc) {
	if (tableStart > dataStart) {
	  tableData = true;
	  tableStart -= dataStart;
	  //fprintf(stderr, "Table in data, offset 0x%lx\n", tableStart);
	}
	else {
	  tableData = false;
	  tableStart -= textStart;
	  //fprintf(stderr, "Table in text, offset 0x%lx\n", tableStart);
	}
      }
      else {
	//fprintf(stderr, "*** Table neither in data nor in text, confused\n");
	//fprintf(stderr, "tableStart 0x%lx, codeOff 0x%lx, codeLen 0x%lx, dataOff 0x%lx, dataLen 0x%lx\n",
	//tableStart, img->codeOffset(), img->codeLength(),
	//img->dataOffset(), img->dataLength());
	
	// Not sure what to do, really. We don't know where it is, and I'm
	// not sure how to check.
      }
      
      for(int i=0;i<maxSwitch;i++){                    
	Address tableEntry = adjustEntry + tableStart + (i * instruction::size());
	//fprintf(stderr, "Table entry at 0x%lx\n", tableEntry);
	if (instructions_->isValidAddress(tableEntry)) {
	  int jumpOffset;
	  if (tableData) {
	  jumpOffset = *((int *)img->getPtrToData(tableEntry));
	  }
	  else
	  jumpOffset = *((int *)instructions_->getPtrToInstruction(tableEntry));
	  
	  //fprintf(stderr, "jumpOffset 0x%lx\n", jumpOffset);
	  Address res = (Address)(jumpStart + jumpOffset);

	  if (img->isCode(res))
	    result += (Address)(jumpStart+jumpOffset);
	  //fprintf(stderr, "Entry of 0x%lx\n", (Address)(jumpStart + jumpOffset));
	}
	else {
	  //fprintf(stderr, "Address not valid!\n");
	}
      }
    }
    else {
      void *ptr;
      if (tableIsRelative) {
	ptr = instructions_->getPtrToInstruction(jumpStartAddress);
	assert(ptr);
	jumpStart = (is64
		     ? *((Address  *)ptr)
		     : *((uint32_t *)ptr));

      }
      ptr = NULL;
      ptr = instructions_->getPtrToInstruction(tableStartAddress);
      assert(ptr);
      tableStart = (is64
		    ? *((Address  *)ptr)
		    : *((uint32_t *)ptr));

      for(int i=0;i<maxSwitch;i++){
      Address tableEntry = adjustEntry + tableStart + (i * instruction::size());
      ptr = instructions_->getPtrToInstruction(tableEntry);
      assert(ptr);
      int jumpOffset = *((int *)ptr);
      result += (Address)(jumpStart+jumpOffset);
      }
    }    
  }
  // No TOC, so we're on Power32 Linux.  Do the ELF thing.
  else
  {
    int entriesAdded = 0;
    for(int i = 0; i < maxSwitch; i++)
    {
      void* ptr = NULL;
      Address tableEntry = tableStartAddress + i*instruction::size();
      if(instructions_->isValidAddress(tableEntry))
      {
	ptr = instructions_->getPtrToInstruction(tableEntry);
      }
      if(ptr)
      {
	int jumpOffset = *((int *)ptr);
	result += (Address)(jumpStartAddress+jumpOffset);
	++entriesAdded;
      }
    }
    if(!entriesAdded)
    {
      setCurrentAddress(initialAddress);
      return false;
    }
    //fprintf(stderr, "Found %d entries in jump table, returning success\n", entriesAdded);
  }

  // Sanity check entries in res
  for (BPatch_Set<Address>::iterator iter = result.begin();
       iter != result.end(); iter++) {
      if ((*iter) % 4) {
          parsing_printf("Warning: found unaligned jump table destination 0x%lx for jump at 0x%lx, disregarding table\n",
                         *iter, initialAddress);
          return false;
      }
  }

  
  setCurrentAddress(initialAddress);
  return true;
}

bool InstrucIter::delayInstructionSupported()
{
  return false;
}

Address InstrucIter::peekPrev()
{
  Address ret = current-instruction::size();
  return ret;
}

Address InstrucIter::peekNext()
{
  Address ret = current + instruction::size();
  return ret;
}

void InstrucIter::setCurrentAddress(Address addr)
{
  current = addr;
  initializeInsn();
}

instruction InstrucIter::getInstruction()
{
  return insn;
}

instruction *InstrucIter::getInsnPtr() {
  instruction *insnPtr = new instruction(insn);
  return insnPtr;
}

// Stack frame creation on Power is easy:
// "stu r1, -120(r1)"
// That being, "store the current value of r1 at *r1 and subtract 120 from r1"
bool InstrucIter::isStackFramePreamble(int &/*unset*/) {
  // We check the entire block. Don't know when it ends,
  // so go until we hit a jump.
  bool foundStackPreamble = false;
  while (instPtr != NULL &&
	 !isAReturnInstruction() &&
	 !isACondBranchInstruction() &&
	 !isACallInstruction() &&
	 !isADynamicCallInstruction() &&
	 !isAJumpInstruction() &&
	 insn.valid()) {
    // We need to bit twiddle.
    instructUnion iu = (*insn);
    if ((iu.dform.op == STUop) &&
	(iu.dform.rt == 1) &&
	(iu.dform.ra == 1)) {
      foundStackPreamble = true;
      break;
    }
    (*this)++;
  }

  return foundStackPreamble;
}

// If we save the return addr, we're not a leaf function.
// Doesn't matter if the function actually makes calls, but
// where to find the return addr.
// Should be "isReturnAddrSave", but that would recompiling
// _everything_.
// As above; looking for "mflr r0", which should be followed
// by a save.

bool InstrucIter::isReturnValueSave() {
  Address currentAddr = current;
  // We check the entire block. Don't know when it ends,
  // so go until we hit a jump.
  bool foundMFLR = false;
  bool foundR0Save = false;
  while (instPtr != NULL &&
	 !isAReturnInstruction() &&
	 !isACondBranchInstruction() &&
	 !isACallInstruction() &&
	 !isADynamicCallInstruction() &&
	 !isAJumpInstruction() &&
	 insn.valid()) {
    // We need to bit twiddle.
    instructUnion iu = (*insn);

    if (iu.raw == MFLR0raw) {
      foundMFLR = true;
    }

    if ((iu.dform.op == STop) &&
	(iu.dform.rt == 0) &&
	(iu.dform.ra == 1)) {
      foundR0Save = true;
      break;
    }
    (*this)++;
  }
  setCurrentAddress(currentAddr);
  return (foundR0Save && foundMFLR);
}

bool InstrucIter::isFrameSetup()
{
  return false;
}

bool InstrucIter::isFramePush()
{
  return false;
}

bool InstrucIter::isALeaveInstruction()
{
  return false;
}

bool InstrucIter::isAnInterruptInstruction()
{
    // TODO: not implemented
    return false;
}

bool InstrucIter::isAnAbortInstruction()
{
  assert(instPtr);
  return !insn.valid();
}


bool InstrucIter::isAnAllocInstruction()
{
  return false;
}

bool InstrucIter::isDelaySlot()
{
  return false;
}

bool InstrucIter::isInterModuleCallSnippet(Address &info) {
  Address currentAddr = current;

  // Template for linkage functions:
  // l      r12,<offset>(r2) // address of call into R12
  // st     r2,20(r1)        // Store old TOC on the stack
  // l      r0,0(r12)        // Address of callee func
  // l      r2,4(r12)        // callee TOC
  // mtctr  0                // We keep the LR static, use the CTR
  // bctr                    // non-saving branch to CTR
    
  bool retval = true;

  for (unsigned i = 0; i < 6; i++) {
    parsing_printf("Checking for linkage at addr 0x%lx\n", current);
    if (!retval) break;

    if (!hasMore()) {
      parsing_printf("failed hasMore check %d: %p, %d, 0x%lx <= 0x%lx <= 0x%lx",
		     i, instPtr, range, base, current, base+range);
      retval = false;
      break;
    }
    instruction scratch = getInstruction();
    parsing_printf("Slot %d: raw bytes 0x%x\n",
		   i, scratch.raw());
    switch(i) {
    case 0: // l r12, <offset> (r2)
      if (((*scratch).dform.op != Lop) ||
	  ((*scratch).dform.rt != 12) ||
	  ((*scratch).dform.ra != 2))  {
	parsing_printf("Insn 0 not load\n");
	retval = false;
      }
      info = (*scratch).dform.d_or_si;
      break;
    case 1: // st     r2,20(r1)
      break; // Don't check this one. 
    case 2: // l      r0,0(r12) 
      if (((*scratch).dform.op != Lop) ||
	  ((*scratch).dform.rt != 0) ||
	  (((*scratch).dform.ra != 1) && ((*scratch).dform.ra != 12))) {
	parsing_printf("Insn 2 not load\n");
	retval = false;
      }
      break;
    case 3: // l      r2,4(r12)
      break;
    case 4: // mtctr  0  
      break; // Could check individually...
    case 5:// bctr 
      if (((*scratch).xlform.op != BCLRop) ||
	  ((*scratch).xlform.xo != BCCTRxop)) {
	parsing_printf("Insn 2 not BCTR\n");
	retval = false;
      }
      break;
    default:
      assert(0);
      break;
    }
    (*this)++;
  }
  setCurrentAddress(currentAddr);
  return retval;

}

void InstrucIter::getAllRegistersUsedAndDefined(std::set<Register> &used,
                                                std::set<Register> &defined) {
    // Register numbers are by the numbering in registerSpace::powerRegisters_t...

    if (isA_MX_Instruction()) {
        // Used and defined, by the comments.
        // TODO: MX register, not MQ... or is it MX?
        used.insert(registerSpace::mq);
        defined.insert(registerSpace::mq);
    }

    // The instrucIter calls tell us if the particular encoding in
    // an instruction is on or off, not which register is used...
    if (isA_RT_ReadInstruction()) {
        // Return of getRTValue is a register number from 0...
        Register reg = getRTValue() + registerSpace::r0;
        used.insert(reg);
    }
    if (isA_RA_ReadInstruction()) {
        used.insert(getRAValue() + registerSpace::r0);
    }
    if (isA_RB_ReadInstruction()) {
        used.insert(getRBValue() + registerSpace::r0);
    }
    if (isA_MRT_ReadInstruction()) {
        // Multiple store - mark all GPRs as used.
        // If we ever use more than r12, increase this.
        // Technically, there's an encoded lower bound... but over-approximate.
        for (unsigned i = registerSpace::r0; i <= registerSpace::r12; i++) {
            used.insert(i);
        }
    }


    if (isA_RT_WriteInstruction()) {
        // Return of getRTValue is a register number from 0...
        Register reg = getRTValue() + registerSpace::r0;
        defined.insert(reg);
    }
    if (isA_RA_WriteInstruction()) {
        defined.insert(getRAValue() + registerSpace::r0);
    }
    if (isA_MRT_WriteInstruction()) {
        // Multiple store - mark all GPRs as used.
        // If we ever use more than r12, increase this.
        // Technically, there's an encoded lower bound... but over-approximate.
        for (unsigned i = registerSpace::r0; i <= registerSpace::r12; i++) {
            used.insert(i);
        }
    }

    //////////////////////////////////
    // FPRs
    //////////////////////////////////

    if (isA_FRT_ReadInstruction()) {
        used.insert(getFRTValue() + registerSpace::fpr0);
    }
    if (isA_FRA_ReadInstruction()) {
        used.insert(getFRAValue() + registerSpace::fpr0);
    }
    if (isA_FRB_ReadInstruction()) {
        used.insert(getFRBValue() + registerSpace::fpr0);
    }
    if (isA_FRC_ReadInstruction()) {
        used.insert(getFRCValue() + registerSpace::fpr0);
    }

    if (isA_FRT_WriteInstruction()) {
        used.insert(getFRTValue() + registerSpace::fpr0);
    }
    if (isA_FRA_WriteInstruction()) {
        used.insert(getFRAValue() + registerSpace::fpr0);
    }

    // SPRs

    std::set<Register> sprsUsed;
    std::set<Register> sprsDefined;

    if (definesSPR(sprsDefined)) {
        std::set<Register> tmp; 
        std::set_union(defined.begin(),
                       defined.end(),
                       sprsDefined.begin(),
                       sprsDefined.end(),
                       insert_iterator<std::set<Register> >(tmp, tmp.begin()) );
        defined = tmp;
    }
    if (usesSPR(sprsUsed)) {
        std::set<Register> tmp;
        std::set_union(used.begin(),
                       used.end(),
                       sprsUsed.begin(),
                       sprsUsed.end(),
                       insert_iterator<std::set<Register> >(tmp, tmp.begin()));
        used = tmp;
    }
}

bool InstrucIter::isSyscall() {
    return ((*insn).iform.op == SVCop);
}

bool InstrucIter::definesSPR(std::set<Register> &sprs) {
    if (((*insn).xfxform.op == MTSPRop) && ((*insn).xfxform.xo == MTSPRxop)) {
        unsigned spr = (*insn).xform.ra + ((*insn).xform.rb << 5);
        switch (spr) {
        case SPR_XER:
            sprs.insert(registerSpace::xer);
            break;
        case SPR_LR:
            sprs.insert(registerSpace::lr);
            break;
        case SPR_CTR:
            sprs.insert(registerSpace::ctr);
            break;
        case SPR_MQ:
            sprs.insert(registerSpace::mq);
            break;
        default:
            break;
        }
        return true;
    }

    // Linking branches define the LR
    if ((((*insn).xlform.op == BCCTRop) && ((*insn).xlform.xo == BCCTRxop) && ((*insn).xlform.lk == 1)) ||
        (((*insn).xlform.op == BCLRop) && ((*insn).xlform.xo == BCLRop) && ((*insn).xlform.lk == 1)) ||
        (((*insn).iform.op == Bop) && ((*insn).iform.lk == 1)) ||
        (((*insn).bform.op == BCop) && ((*insn).bform.lk == 1))) {
        sprs.insert(registerSpace::lr);
    }

    return false;
}


bool InstrucIter::usesSPR(std::set<Register> &sprs) {
    if (((*insn).xfxform.op == MFSPRop) && ((*insn).xfxform.xo == MFSPRxop)) {
        unsigned spr = (*insn).xform.ra + ((*insn).xform.rb << 5);
        switch (spr) {
        case SPR_XER:
            sprs.insert(registerSpace::xer);
            break;
        case SPR_LR:
            sprs.insert(registerSpace::lr);
            break;
        case SPR_CTR:
            sprs.insert(registerSpace::ctr);
            break;
        case SPR_MQ:
            sprs.insert(registerSpace::mq);
            break;
        default:
            break;
        }
        return true;
    }

    // Branches to SPRs
    if (((*insn).xlform.op == BCCTRop) && ((*insn).xlform.xo == BCCTRxop)) {
        sprs.insert(registerSpace::ctr);
        return true;
    }
    if (((*insn).xlform.op == BCLRop) && ((*insn).xlform.xo == BCLRxop)) {
        sprs.insert(registerSpace::lr);
        return true;
    }
    
    return false;
}

bool InstrucIter::isANopInstruction()
{
   return false;
}
